{:
PURPOSE AND IMPLEMENTATION:
	This unit defines some useful string handling routines.

HISTORY:
	??/??/96 - Created by Colin Patrick Sarsfield (colin.sarsfield@usa.net)
	11/08/96 - Left and Right functions added by CPS
	05/21/97 - IntlStrToFloat added by CPS
  11/28/01 - Fixed bug in CompareStrNumMix where equal values did not result in
             zero return value

COPYRIGHT:
	Released into the public domain.
}
unit Strlib;

interface

uses
	Classes;

type
  CharSet = set of Char;
  TArrayOfString = array of String;

const
  NL = #13#10;
  TB = #9;

function AnsiLowerCaseNonAcronyms(s: String): String;
function AnsiTitleCaseNonAcronyms(s: String): String;
function AnsiTitleCase(s: String): String;
function BoolToStr(const b: Boolean): string;
function CompareStrNumMix(const str1, str2: String): Integer;
function CountChar(const s: String; c: Char): Integer;
function DequoteString(s: String): String;
function DescToFileName(Desc: String; ReplaceChar: Char): String;
procedure DivideStrInt(const Str: string; var BegStr: string; var Int: Integer);
function DropLeadingZeros(S: String): String;
function DropTrailingZeros(const S: String): String;
procedure FileSplit(const Path: String; var Dir, Name, Ext: String);
function FloatToFixedStr(const Value: Extended; Precision: Integer): String;
function ForgivingStrToDate(const DateStr: String): TDateTime;
function ForgivingTryStrToCurr(Str: String; out Curr: Currency): Boolean;
function FormatAsIdentifier(const AString: String): String;
function FracToStr(const n, d: Integer): string;
function GetStringBetween(s, Before, After: String): String;
function Hash(const S: string): Integer;
function IntlStrToFloat(const AString: string): Single;
function IntToAlpha(const i: Longint): string;
function IsInList(const Value, List: string): Boolean;
function Join(Strs: TArrayOfString; Delim: String): String; overload;
function Join(Ints: array of Integer; Delim: String): String; overload;
function Left(const AString: string; i: Integer): string;
function LookupGroup(const Number: string; const Groups: array of string): string;
function LookupOrdStr(const s: String; const a: array of String): Integer;
function LTrim(const AString: string): string;
function LTrimChars(const AString: string; TrimChars: CharSet): string;
procedure MergeValues(A, B: TStringList);
function NextStrColumn(S: String; var Pos: Integer): String;
function PhraseToAcronym(S: String): String;
function PosAfter(SubStr, S: String; StartingAt: Integer): Integer;
function QtyList(A: TStringList): String;
function RepeatStr(const Str: String; const qty: Integer): String;
function Replace(const PTarget, BeforeStr, AfterStr: string): string;
function Right(const AString: string; i: Integer): string;
function RScan(const AString: string; chr: Char): Integer;
function RTrim(const AString: string): string;
function RTrimChars(const AString: string; TrimChars: CharSet): string;
function Scan(const AString: string; chr: Char): Integer;
{$IFNDEF WIN32}
procedure SetLength(var S: string; NewLength: Integer);
{$ENDIF}
function Split(const S, Delim: String): TArrayOfString;
function SplitString(var Pos: Integer; s, Delim: String): String;
procedure StreamToString(AStream: TStream; var AString: string);
function StringOfChar(const AChar: Char; Length: Integer): string;
procedure StringToStream(const AString: string; AStream: TStream);
function TerminateSentence(const AString: String): String;
function Trim(const AString: string): string;
function TrimChars(const AString: String; TrimChars: CharSet): string;
function WordToInt(const Word: String): Integer;

implementation

uses
  Windows, SysUtils, DateUtils, Math, StrUtils;
    
type
  TStringFormatFunc = function(s: String): String;

function AnsiAdjustCaseNonAcronyms(s: String; FormatFunc: TStringFormatFunc): String;
var
  i, UpperCount, WordStart: Integer;
begin
  UpperCount := 0;
  WordStart := 1;
  Result := '';
  for i := 1 to Succ(Length(s)) do begin
    if (i <= Length(s))
        and (IsCharUpper(s[i]) or IsCharLower(s[i]) or (s[i] = '-')) then begin
      if WordStart = 0 then WordStart := i;
      if IsCharUpper(s[i]) then
        Inc(UpperCount);
    end else begin
      if WordStart <> 0 then begin
        if ((UpperCount = 1) and IsCharUpper(s[WordStart]))
            or (UpperCount = 0) then
          Result := Result + FormatFunc(Copy(s, WordStart, i - WordStart))
        else
          Result := Result + Copy(s, WordStart, i - WordStart);
        UpperCount := 0;
        WordStart := 0;
      end;
      if i <= Length(s) then
        Result := Result + s[i];
    end;
  end;
end;

{:Returns the string lowercased, except for words that appear to be acronyms.
Note that this function does nothing to uppercase strings.}
function AnsiLowerCaseNonAcronyms(s: String): String;
begin
  Result := AnsiAdjustCaseNonAcronyms(s, @AnsiLowerCase);
end;

{:Returns the string title cased, except for words that appear to be acronyms.
Note that this function does nothing to uppercase strings.}
function AnsiTitleCaseNonAcronyms(s: String): String;
begin
  Result := AnsiAdjustCaseNonAcronyms(s, @AnsiTitleCase);
end;

{:Converts the given string to title case, not changing any letters already
capitalized.}
function AnsiTitleCase(s: String): String;
var
  i: Integer;
begin
  for i := 1 to Length(s) do
    if (i = 1) or (s[Pred(i)] = ' ') then
      s[i] := AnsiUpperCase(s[i])[1];
  Result := s;
end;

{:Return a string representation for the boolean value b.  (e.g. True = 'True') }
function BoolToStr(const b: Boolean): string;
begin
  if b then
    Result := 'True'
  else
    Result := 'False';
end;

{:compare two strings.  Unlike a normal string comparison, digits 0..9 are seen
as a number with any other adjoining numbers and these numbers are compared as
whole numbers.  An example of a list properly sorted using this comparision
function:

AH-1
AH-2b
AH-12-15c
AH-31

This function is case-insensitive.}
function CompareStrNumMix(const str1, str2: String): Integer;
var
  w1, w2: String;
  p1, p2: Integer;
  isdigit1, isdigit2: Boolean;
const
  Digits = ['0'..'9'];

  procedure GetDigits(const source: String; var work: String; var pos: Integer);
  begin
    while (pos <= Length(source)) and (source[pos] in Digits) do begin
      work := work + source[pos];
      Inc(pos)
    end;
  end;
  
  procedure GetString(const source: String; var work: String; var pos: Integer);
  begin
    while (pos <= Length(source)) and not (source[pos] in Digits) do begin
      work := work + source[pos];
      Inc(pos)
    end;
  end;
begin
  // if str1 < str2 then negative
  // if str1 > str2 then positive
  // if str1 = str2 then zero
  p1 := 1;
  p2 := 1;
  repeat
    { terminate with longer string being greater if end of either string is
      reached }
    if p1 > Length(str1) then begin
      if p2 > Length(str2) then
        Result := 0
      else
        Result := -1;
      Exit
    end;
    if p2 > Length(str2) then begin Result := 1; Exit end;
    w1 := '';
    w2 := '';

    isdigit1 := str1[p1] in Digits;
    isdigit2 := str2[p2] in Digits;
    { if different types of strings then exit }
    if isdigit1 <> isdigit2 then begin Result := Ord(isdigit1) - Ord(isdigit2); Exit end;
    if isdigit1 then begin
      { number portion }
      GetDigits(str1, w1, p1);
      GetDigits(str2, w2, p2);
      try
        Result := StrToInt(w1) - StrToInt(w2);
      except
        // if StrToInt fails, just compare the text normally
        Result := CompareText(w1, w2);
      end;
    end
    else begin
      // string portion
      GetString(str1, w1, p1);
      GetString(str2, w2, p2);
      Result := CompareText(w1, w2);
    end;
  until Result <> 0;
end;

{:Returns the occurrences of c in s.}
function CountChar(const s: String; c: Char): Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := 1 to Length(s) do
    if s[i] = c then Inc(Result);
end;

{:Revert a quoted-style string to normal.  Leave other styles of strings
untouched.  For example,

"1"" pipe thread"

becomes

1" pipe thread
}
function DequoteString(s: String): String;
var
  Len: Integer;
begin
  Len := Length(s);
  if ((Copy(s, 1, 1) = '"') and (Copy(s, Len, 1) = '"')) then begin
    s := Copy(s, 2, Len - 2);
    s := Replace(s, '""', '"');
  end;
  Result := s;
end;

{:Convert illegal filename characters, non-ASCII characters with ReplaceChar.
}
function DescToFileName(Desc: String; ReplaceChar: Char): String;
var
  i: Integer;
begin
  Result := Desc;
  for i := 1 to Length(Desc) do
    if (Desc[i] in ['\','/',':','*','?','"','<','>','|'])
        or (Byte(Desc[i]) > 127) then
      Result[i] := ReplaceChar;
end;

{:Assuming Str begins with non-numeric characters and ends with numeric
characters (i.e., 0..9), return the non-numeric portion in BegStr and convert
the numeric portion to an integer returning it in Int.}
procedure DivideStrInt(const Str: string; var BegStr: string; var Int: Integer);
var
	TStr: string;
	i: Integer;
begin
  for i := Length(Str) downto 1 do
    if (Str[i] < '0') or (Str[i] > '9') then
      Break;
  BegStr := Copy(Str, 1, i);
  if i = Length(Str) then
    Int := 0
  else begin
    TStr := Copy(Str, i + 1, Length(Str) - i);
    try { capture range check errors, etc. }
      Int := StrToInt(TStr);
    except
      Int := 0;
    end;
	end;
end;

{:Eliminates any leading zeros from the given string; however, leaves one zero
if the original string was all zeroes.}
function DropLeadingZeros(S: String): String; 
var
  i: Integer;
begin
  S := Trim(S);
  for i := 1 to Length(S) do
    if S[i] <> '0' then begin
      Result := Copy(S, i, MaxInt);
      Exit;
    end;
  if Length(S) > 0 then
    Result := '0'
  else
    Result := '';
end;

{:Eliminates .00 from end of whole numbers.  (i.e., 10.00 = 10, 10.50 = 10.5,
10.05 = 10.05)  Looks through string for decimal.  If found, goes to the end
and looks back one character for a decimal or zero.  Continues that action
eliminating zeros and decimals until whole number is found to left of decimal.}
function DropTrailingZeros(const S: String): String;
var
  I: Integer;
  L: Integer;
begin
  Result := S;
  I := POS(DecimalSeparator,Result);
  if I <> 0 then begin
    L := Length(Result);
    while Result[L] = '0' do
      dec(L);
    if Result[L] = DecimalSeparator then
      dec(L);
    SetLength(Result,L)
  end;
end;

{:Split a file path "Path" into three parts:  its directory "Dir", its name
"Name", and its extension "Ext".}
procedure FileSplit(const Path: String; var Dir, Name, Ext: String);
var
	NamePos, ExtPos: Integer;
begin
	NamePos := RScan(Path, '\');
	if NamePos = 0 then NamePos := RScan(Path, ':');
	if NamePos = 0 then NamePos := 1 else Inc(NamePos);
	ExtPos := RScan(Path, '.');
	if ExtPos < NamePos then ExtPos := Length(Name);
	Dir := Copy(Path, 1, NamePos - 1);
	Name := Copy(Path, NamePos, ExtPos - NamePos);
	Ext := Copy(Path, ExtPos, Length(Path) - ExtPos + 1);
end;

{:Converts a floating point value to a string, respecting the precision
parameter.  The resulting string will never use scientific format.  Any
trailing zeros to the right of the decimal point will be dropped from
the string.}
function FloatToFixedStr(const Value: Extended; Precision: Integer): String;
var
  Exponent, ExponentPos: Integer;
  Negative: Boolean;
begin
  Result := FloatToStrF(Value, ffExponent, Precision, 4);
  ExponentPos := Pos('E', Result);
  Exponent := StrToInt(Copy(Result, Succ(ExponentPos), MaxInt));
  Negative := Left(Result, 1) = '-';
  Delete(Result, ExponentPos, MaxInt);
  if Negative then Delete(Result, 1, 1);
  Delete(Result, 2, 1); // delete decimal point

  if Exponent >= Length(Result) then
    Result := Result + RepeatStr('0', Exponent - Length(Result) + 1)
  else if Exponent < 0 then
    Result := '0' + DecimalSeparator + RepeatStr('0', -1 - Exponent) + Result
  else if Exponent < Pred(Length(Result)) then
    Insert(DecimalSeparator, Result, 2 + Exponent);
  Result := DropTrailingZeros(Result);
  if Negative then Insert('-', Result, 1);
end;

{:Convert a string to a date, accepting a wide range of formats}
function ForgivingStrToDate(const DateStr: String): TDateTime;
type
  TDatePart = (dpDay, dpMonth, dpYear);
var
  i, PartIdx, DayPos, YearPos, MonthPos: Integer;
  Parts: array[0..2] of String;
  Day, Year, Month, CenturyBase, CurrentYear: Word;

  function GetDateOrder: Char;
  var
    m: Integer;
  begin
    Result := 'M';
    for m := 1 to Length(ShortDateFormat) do begin
      Result := UpperCase(ShortDateFormat[m])[1];
      if Pos(Result, 'YMD') <> 0 then Exit;
    end;
  end;

  function PartToInt(DatePart: String; out IsMonth, IsValid: Boolean): Integer;
  const
    MonthAbbrevs: array[1..12] of String[9] = (
      'JANUARY', 'FEBRUARY', 'MARCH', 'APRIL', 'MAY', 'JUNE', 'JULY', 'AUGUST', 'SEPTEMBER', 'OCTOBER', 'NOVEMBER', 'DECEMBER'
      );
  var
    m: Integer;
  begin
    IsValid := (DatePart <> '') and (DatePart[1] in ['0'..'9']) and TryStrToInt(DatePart, Result);
    IsMonth := not IsValid;
    if IsMonth and (DatePart <> '') then begin
      DatePart := UpperCase(DatePart);
      for m := Low(MonthAbbrevs) to High(MonthAbbrevs) do
        if DatePart = Left(MonthAbbrevs[m], Length(DatePart)) then begin
          IsValid := True;
          Result := m;
          Break;
        end;
    end;
  end;

  function SimplePartToInt(DatePart: String): Integer;
  var
    Junk1, Junk2: Boolean;
  begin
    Result := PartToInt(DatePart, Junk1, Junk2);
  end;

  function ValidPart(DatePart: String): Boolean;
  var
    Junk: Boolean;
  begin
    PartToInt(DatePart, Junk, Result);
  end;

  function IsMonth(TestStr: String): Boolean;
  var
    m: Integer;
  begin
    m := SimplePartToInt(TestStr);
    Result := (m >= 1) and (m <= 12);
  end;

  function IsYear(TestStr: String): Boolean;
  var
    IsMonth, Junk: Boolean;
  begin
    PartToInt(TestStr, IsMonth, Junk);
    Result := not IsMonth;
  end;

  function IsDay(TestStr: String): Boolean;
  var
    d: Integer;
    IsMonth, Junk: Boolean;
  begin
    d := PartToInt(TestStr, IsMonth, Junk);
    Result := not IsMonth and (d >= 1) and (d <= 31);
  end;

  procedure Find(var Pos: Integer; DatePart: TDatePart);
  var
    m: Integer;
    Found: Boolean;
  begin
    Found := False;
    for m := Low(Parts) to High(Parts) do begin
      if (m = DayPos) or (m = MonthPos) or (m = YearPos) then Continue;
      case DatePart of
      dpDay: Found := IsDay(Parts[m]);
      dpMonth:
        if IsMonth(Parts[m]) then begin
          if Pos = -1 then Pos := m;
          if not IsDay(Parts[m]) then Found := True;
        end;
      dpYear: Found := IsYear(Parts[m]);
      end;
      if Found then begin
        Pos := m;
        Break;
      end;
    end;
  end;
begin
  if TryStrToDate(DateStr, Result) then Exit;

  PartIdx := 0;
  for i := Low(Parts) to High(Parts) do
    Parts[i] := '';

  for i := 1 to Length(DateStr) do begin
    if ValidPart(Parts[PartIdx] + DateStr[i]) then
      Parts[PartIdx] := Parts[PartIdx] + DateStr[i]
    else if ValidPart(Parts[PartIdx]) then begin
      Inc(PartIdx);
      if PartIdx > High(Parts) then Break;
      if ValidPart(Parts[PartIdx] + DateStr[i]) then
        Parts[PartIdx] := Parts[PartIdx] + DateStr[i]
    end;
  end;

  case GetDateOrder of
  'Y': begin YearPos := 0; MonthPos := 1; DayPos := 2; end;
  'D': begin YearPos := 2; MonthPos := 1; DayPos := 0; end;
  else begin YearPos := 2; MonthPos := 0; DayPos := 1; end;
  end;

  if not (IsYear(Parts[YearPos]) and IsMonth(Parts[MonthPos]) and IsDay(Parts[DayPos])) then begin
    YearPos := -1; MonthPos := -1; DayPos := -1;

    Find(MonthPos, dpMonth);
    Find(DayPos, dpDay);
    Find(YearPos, dpYear);

    if (YearPos = -1) and (MonthPos = -1) and (DayPos = -1) then begin
      Result := 0;
      Exit;
    end;
  end;
  DecodeDate(Date, Year, Month, Day);
  CurrentYear := Year;
  Day := 1; // default to first day of month if unspecified

  if YearPos <> -1 then
    Year := SimplePartToInt(Parts[YearPos]);
  if MonthPos <> -1 then
    Month := SimplePartToInt(Parts[MonthPos]);
  if DayPos <> -1 then
    Day := SimplePartToInt(Parts[DayPos]);

  if Year < 100 then begin
    CenturyBase := CurrentYear - TwoDigitYearCenturyWindow;
    Inc(Year, CenturyBase div 100 * 100);
    if (TwoDigitYearCenturyWindow > 0) and (Year < CenturyBase) then
      Inc(Year, 100);
  end;

  Day := Min(Day, DaysInAMonth(Year, Month));
  Result := EncodeDate(Year, Month, Day);
end;

{:Improves on the functionality of TryStrToCurr by adding support for ignoring
thousand separators and dollar signs.}
function ForgivingTryStrToCurr(Str: String; out Curr: Currency): Boolean;
begin
  Str := Replace(Str, '$', '');
  Str := Replace(Str, ThousandSeparator, '');
  Result := TryStrToCurr(Str, Curr);
end;

{:Take a string and format it as an identifier (e.g.,
"Cond. fluid" = "CondFluid"
"24 times" = "Times"
"Cond. fluid_Water" = "CondFluid_Water"
}
function FormatAsIdentifier(const AString: String): String;
var
  i, Len: Integer;
  Cap: Boolean;
begin
  Len := Length(AString);
  Result := '';
  i := 1;
  // skip over junk at beginning
  while (i <= Len) and not (AString[i] in ['a'..'z', 'A'..'Z', '_']) do
    Inc(i);
  Cap := True;
  while i <= Len do begin
    if AString[i] = '_' then begin
      Result := Result + '_';
      Cap := True
    end
    else if AString[i] in ['a'..'z', 'A'..'Z', '0'..'9'] then begin
      if Cap then begin
        Result := Result + UpperCase(AString[i]);
        Cap := False;
      end
      else
        Result := Result + AString[i];
    end
    else
      Cap := True;
    Inc(i)
  end;
end;

{:Convert a fraction to its string representation where "n" is the numerator
and "d" is the denomiator.  Simplify the fraction if possible.  Fractions will
be displayed as an integer (optionally) followed by a fraction less than one
(optionally).  Note that this function only works for positive integers or zero. }
function FracToStr(const n, d: Integer): string;
var
  i, w, nn, dd: Integer;
begin
  // fail if n or d are negative
  if (n < 0) or (d < 0) then
    raise Exception.Create('FracToStr does not support negative integers');
  // calculate the whole number portion of the fraction
  w := n div d;
  // calculate the remaining numerator after the whole number is subtracted out
  nn := n mod d;
  // store a copy of the denominator for simplification
	dd := d;
  // check all possible divisors to simply
  for i := d - 1 downto 2 do
    // if the nominator and the denominator are both divisable by the same
    // number than divide by that number to simplify.
    if ((dd mod i) = 0) and ((nn mod i) = 0) then begin
      dd := dd div i;
      nn := nn div i;
    end;
  // if the whole number > 0 then represent it in the result
  if w > 0 then
    Result := Format('%d ', [w])
	else // otherwise blank out the result
    Result := '';
  // if the numerator is also zero than display zero
  if nn = 0 then begin
    if w = 0 then Result := '0';
  end
  else // otherwise add the string representation of the fraction to the result
    Result := Result + Format('%d/%d', [nn, dd]);
end;

{:Returns the text occurring between two given substrings, or '' if not found.}
function GetStringBetween(s, Before, After: String): String;
var
  TextStart, BeforePos, AfterPos: Integer;
begin
  Result := '';
  BeforePos := Pos(Before, s);
  if BeforePos > 0 then begin
    TextStart := BeforePos + Length(Before);
    AfterPos := PosAfter(After, s, TextStart);
    if AfterPos > 0 then
      Result := Copy(s, TextStart, AfterPos - TextStart);
  end;
end;

{:Returns a hash for the string S.}
function Hash(const S: string): Integer;
var
  I: Integer;
  PC1, PC2: PChar;
  P1, P2: PInteger;
begin
  Result := 0;
  I := Length(S);
  P1 := PInteger(pointer(@S[1]));
  PC2 := @S[I+1];
  P2 := PInteger(integer(pc2) and -3);
  while P1 <> P2 do begin
    Result := Result * 988053892081
        + (P1^ and 255) * 991026973
        + ((P1^ shr 8) and 255) * 994009
        + ((P1^ shr 16) and 255) * 997
        + (P1^ shr 24);
    Inc(P1);
  end;
  PC1 := PChar(P1);
  while PC1 <> PC2 do begin
    Result := Result * 997 + Ord(PC1^);
    Inc(PC1);
  end;
end;

{:This function will convert a String to a Single whether the decimal seperator
used is the localized (international) decimal seperator or the period (i.e., '.'). }
function IntlStrToFloat(const AString: string): Single;
var
  RCode: Integer;
begin
  try
    // StrToFloat recognizes the international decimal seperator (e.g., '.' in
    // USA, ',' in parts of Latin America)
    Result := StrToFloat(AString);
  except
    // Val always recognizes '.' as the decimal seperator
    Val(AString, Result, RCode);
		if RCode <> 0 then
			raise EConvertError.Create('Unable to convert string to number: ''' + AString + '');
  end;
end;

{:Convert an integer to a base 26 number based on the alphabet much like
column headings in most spreadsheet programs (e.g., A, B, C .., Z, AA, AB, etc.).}
function IntToAlpha(const i: Longint): string;
begin
	if i <= 26 then
    Result := Chr(Ord('A') + i - 1)
  else
    Result := IntToAlpha(i div 26) + Chr(Ord('A') + (i mod 26));
end;

{:This function returns True if Value is in comma-delimited List. }
function IsInList(const Value, List: string): Boolean;
begin
  Result := (Value = List)
    or (Pos(',' + Value + ',', List) <> 0)
    or (Right(List, Length(Value) + 1) = (',' + Value))
    or (Left(List, Length(Value) + 1) = (Value + ','))
end;

{:This function joins together a dynamic array of strings with a delimiter,
and returns the composite string.}
function Join(Strs: TArrayOfString; Delim: String): String;
var
  i, UsedLength, StrIdx: Integer;

  procedure AddStr(const s: String);
  begin
    if S <> '' then begin
      StrCopy(PChar(@Result[StrIdx]), PChar(s));
      Inc(StrIdx, Length(s));
    end;
  end;
begin
  UsedLength := Pred(Length(Strs)) * Length(Delim);
  for i := Low(Strs) to High(Strs) do
    Inc(UsedLength, Length(Strs[i]));
  SetLength(Result, UsedLength);

  StrIdx := 1;
  for i := Low(Strs) to High(Strs) do begin
    AddStr(Strs[i]);
    if i <> High(Strs) then AddStr(Delim);
  end;
end;

{:This function joins together a dynamic array of integers with a delimiter,
and returns the composite string.}
function Join(Ints: array of Integer; Delim: String): String; overload;
var
  i: Integer;
  Strs: TArrayOfString;
begin
  SetLength(Strs, Length(Ints));
  for i := Low(Ints) to High(Ints) do
    Strs[i] := IntToStr(Ints[i]);
  Result := Join(Strs, Delim);
end;

{:This function returns a string "i" characters long starting at the left
(beginning) of "AString".  Note that if "AString" is less that "i" characters
long this function will return the whole string. }
function Left(const AString: string; i: Integer): string;
begin
  if i >= Length(AString) then
    Result := AString
  else
  	Result := Copy(AString, 1, i);
end;

{:This function returns the string in Groups that returns True when passed to
IsInList with Number.}
function LookupGroup(const Number: string; const Groups: array of string): string;
var
  i: Integer;
begin
  for i := Low(Groups) to High(Groups) do
    if IsInList(Number, Groups[i]) then begin
      Result := Groups[i];
      Exit;
    end;
  Result := '';
end;

{:This function returns an integer indicating the position of s in a.  Note that
this is often used to look up an ordinal based on its string representation;
hence its name.}
function LookupOrdStr(const s: String; const a: array of String): Integer;
var
  i: Integer;
begin
  for i := Low(a) to High(a) do
    if a[i] = s then begin
      Result := i;
      Exit;
    end;
  raise Exception.CreateFmt('Unable to lookup ordinal string value: "%s"', [s]);
end;

{:This function returns the remainder of "AString" after any leading spaces are
removed.}
function LTrim(const AString: string): string;
var
  i, b, l: Integer;
begin
  b := 1;
	l := Length(AString);
	for i := 1 to l do
		if AString[i] = ' ' then
      b := i + 1
    else
      Break;
  if b > l then
    Result := ''
  else
		Result := Copy(AString, b, l - b + 1)
end;

{:This function strips off leading characters found in TrimChars and
returns the resulting string.}
function LTrimChars(const AString: string; TrimChars: CharSet): string;
var
  i, b, e: Integer;
begin
  b := 1;
  e := Length(AString);
  for i := 1 to e do
    if AString[i] in TrimChars then
      b := i + 1
    else
      Break;
  if b > e then
    Result := ''
  else
    Result := Copy(AString, b, e - b + 1)
end;

{:This procedure takes two TStringList's populated with lines like "coil=true"
(reference the "Values" property in the help) and merges them.  The result is
placed in A.  Note that for any strings that both A & B have B have the part to
the left of the equals sign in common the merged list will use the line from B.}
procedure MergeValues(A, B: TStringList);
var
	i, p: Integer;
begin
	for i := 0 to Pred(B.Count) do begin
		p := Pos('=', B[i]);
		if (p <> 0) then
			A.Values[Copy(B[i], 1, p - 1)] := Copy(B[i], p + 1, Length(B[i]) - p);
	end;
end;

{:This function extracts the next "column" out of "S" which should be a line
from either a space or tab delimited text file.  Note that Pos should be 1 on
the first call to this function for a given string and the Pos returned by this
function should be used in subsequent calls.}
function NextStrColumn(S: String; var Pos: Integer): String;
var
  Start, Len: Integer;

  function IsSpace(C: Char): Boolean;
  begin
    Result := C in [' ', #9];
  end;
begin
  Len := Length(S);
  while IsSpace(S[Pos]) and (Pos <= Len) do Inc(Pos);
  if Pos > Len then begin Result := ''; Exit end;
  Start := Pos;
  while (Pos <= Len) and not IsSpace(S[Pos]) do Inc(Pos);
  Result := Copy(S, Start, Pos - Start);
end;

{:Converts a phrase to an acronym, dropping common words like "of", and replacing
some words with shorthand (e.g. "number" becomes "#").  }
function PhraseToAcronym(S: String): String;
const
  WordsToIgnore = 'a,the,to,for,from,in,at,with,toward,of';
  WordsToReplace: array[0..0, 0..1] of String = (
    ('number', '#')
    );
var
  i, j: Integer;
  Words: TArrayOfString;
begin
  Words := nil; // to make compiler happy
  
  if S = UpperCase(S) then begin
    // String appears to be an acronym already.
    Result := S;
    Exit;
  end;

  Words := Split(S, ' ');
  Result := '';
  for i := Low(Words) to High(Words) do begin
    for j := Low(WordsToReplace) to High(WordsToReplace) do
      if CompareText(WordsToReplace[j, 0], Words[i]) = 0 then begin
        Words[i] := '';
        Result := Result + WordsToReplace[j, 1];
      end;

    if IsInList(Words[i], WordsToIgnore) then Continue;
    if Words[i] = '' then Continue;

    if not IsCharAlpha(Words[i][1]) then Continue;

    Result := Result + UpperCase(Words[i][1]);
  end;
end;

{:This function is just like the "Pos" function that is provided with Delphi
except it accepts an additional argument, "StartingAt".  This argument tells
this function where to start its search whereas the "Pos" function always starts
at position 1.}
function PosAfter(SubStr, S: String; StartingAt: Integer): Integer;
begin
	if StartingAt < 1 then StartingAt := 1;
	Delete(S, 1, StartingAt - 1);
	Result := Pos(SubStr, S);
	if Result <> 0 then
		Result := Result + StartingAt - 1;
end;

{:this function takes a TStringList and formats it as a "quantity list" of the
string list's values.  For example:
The StringList 'A','B','A','B','B','C'
becomes '(2)A/(3)B/(1)C'
If the string list contains only one item it is considered a special case and
that item is returned without any change in formatting.
Note: the string list is sorted by list function}
function QtyList(A: TStringList): String;
var
  i, qty: Integer;
  str: String;
begin
  if A.Count = 1 then begin
    Result := A[0];
    Exit;
  end;
  A.Sort;
  i := 0;
  Result := '';
  while i < A.Count do begin
    str := A[i];
    qty := 1;
    while (i < Pred(A.Count)) and (A[i + 1] = str) do begin
      Inc(qty);
      Inc(i);
    end;
    if Result <> '' then Result := Result + '/';
    Result := Result + Format('(%d)%s', [qty, str]);
    Inc(i)
  end;
end;

{:This function creates a string that is the concatenation of "qty" instances of
the string "Str".}
function RepeatStr(const Str: String; const qty: Integer): String;
var
  i: Integer;
begin
  if qty > 0 then begin
    Result := Str;
    for i := 2 to qty do
      Result := Result + Str;
  end
  else
    Result := '';
end;

{:This function returns a modified version of "PTarget" where all instances of
"BeforeStr" have been replaced by "AfterStr".}
function Replace(const PTarget, BeforeStr, AfterStr: string): string;
var
	i, j: Integer;
begin
	Result := PTarget;
	{ start at beginning of string }
  i := 1;
  while True do begin
    j := i;
    i := Pos(BeforeStr, Copy(Result, i, Length(Result) - i + 1));
    if i = 0 then
      Break;
    Inc(i, j - 1);
    Delete(Result, i, Length(BeforeStr));
    Insert(AfterStr, Result, i);
    { adjust i to correct position }
    i := i + Length(AfterStr);
  end;
  Result := Result;
end;

{:This function returns a string "i" characters long starting at the right
(end) of "AString".  Note that if "AString" is less that "i" characters
long this function will return the whole string. }
function Right(const AString: string; i: Integer): string;
begin
  if i >= Length(AString) then
    Result := AString
  else
    Result := Copy(AString, Length(AString) - i + 1, i);
end;

{:This function returns the index of the last instance of "chr" in "AString".
It returns 0 if none was found.}
function RScan(const AString: string; chr: Char): Integer;
var
	i: Integer;
begin
	Result := 0;
	for i := Length(AString) downto 1 do
		if AString[i] = chr then begin
			Result := i;
			Exit;
		end;
end;

{:This function returns the remainder of "AString" after any trailing spaces are
removed.}
function RTrim(const AString: string): string;
var
  i, e: Integer;
begin
  e := Length(AString);
  for i := Length(AString) downto 1 do
    if AString[i] = ' ' then
      e := i - 1
    else
      Break;
  if 1 > e then
    Result := ''
  else
    Result := Copy(AString, 1, e)
end;

{:This function strips off trailing characters found in TrimChars and
returns the resulting string.}
function RTrimChars(const AString: string; TrimChars: CharSet): string;
var
  i, e: Integer;
begin
  e := Length(AString);
  for i := e downto 1 do
    if AString[i] in TrimChars then
      e := i - 1
    else
      Break;
  if 1 > e then
    Result := ''
  else
    Result := Copy(AString, 1, e)
end;

{:This function returns the index of the first instance of "chr" in "AString".
It returns 0 if none was found.}
function Scan(const AString: string; chr: Char): Integer;
var
	i: Integer;
begin
	Result := 0;
	for i := 1 to Length(AString) do
		if AString[i] = chr then begin
			Result := i;
			Exit;
		end;
end;

{$IFNDEF WIN32}

{:This function is to enable portability between 16-bit and 32-bit programs.
Since 32-bit Delphi defines a SetLength function, this defines one for 16-bit
programs.}
procedure SetLength(var S: string; NewLength: Integer);
begin
	S[0] := Char(NewLength);
end;
{$ENDIF}

procedure TestSplitString;
const
  s = 'first,second,third,,fourth,';
var
  Pos: Integer;
begin
  Pos := 1;
  Assert(SplitString(Pos, s, ',') = 'first');
  Assert(SplitString(Pos, s, ',') = 'second');
  Assert(SplitString(Pos, s, ',') = 'third');
  Assert(SplitString(Pos, s, ',') = '');
  Assert(SplitString(Pos, s, ',') = 'fourth');
  Assert(SplitString(Pos, s, ',') = '');
  Assert(Pos = 0);

  Pos := 1;
  Assert(SplitString(Pos, '', ',') = '');
  Assert(Pos = 0);

  WriteLn('SplitString test completed.');
end;

{:This function splits apart delimited substrings based on the token Delim.
The result is a dynamic array of strings.}
function Split(const S, Delim: String): TArrayOfString;
var
  i, Capacity, Pos: Integer;
begin
  if S = '' then begin
    SetLength(Result, 0);
    Exit;
  end;

  i := 0;
  Pos := 1;
  Capacity := 16;
  SetLength(Result, Capacity);
  repeat
    if i >= Capacity then begin
      Capacity := Capacity * 2;
      SetLength(Result, Capacity);
    end;
    Result[i] := SplitString(Pos, S, Delim);
    Inc(i);
  until Pos = 0;
  SetLength(Result, i);
end;

{:This function splits apart delimited substrings based on the token Delim.
Each call returns one substring and adjusts Pos to the start of the next
substring, or 0 if this is the last substring. }
function SplitString(var Pos: Integer; s, Delim: String): String;
var
  NewPos: Integer;
begin
  if (Pos < 1) or (Pos > Length(s)) then begin
    Result := '';
    Pos := 0;
    Exit;
  end;

  NewPos := PosAfter(Delim, s, Pos);
  if NewPos = 0 then begin
    Result := Copy(s, Pos, Length(s) - Pos + 1);
    Pos := 0;
  end else begin
    Result := Copy(s, Pos, NewPos - Pos);
    Pos := NewPos + Length(Delim);
  end;
end;

{:This procedure duplicates the contents of "AStream" and places it in "AString".}
procedure StreamToString(AStream: TStream; var AString: string);
var
  L: Integer;
begin
  AStream.Read(L, Sizeof(Integer));
  SetLength(AString, L);
  if L > 0 then
    AStream.Read(AString[1], L);
end;

{:This function returns a string "Length" characters long where each character
is "AChar".}
function StringOfChar(const AChar: Char; Length: Integer): string;
begin
  SetLength(Result, Length);
  if Length > 0 then
    FillChar(Result[1], Length, AChar);
end;

{:This procedure writes the contents of "AString" to "AStream".}
procedure StringToStream(const AString: string; AStream: TStream);
var
  L: Integer;
begin
  L := Length(AString);
  AStream.Write(L, Sizeof(Integer));
  if L > 0 then
    AStream.Write(AString[1], L);
end;

{:This function returns the given string, with a period appended if the sentence
has no terminating punctuation.}
function TerminateSentence(const AString: String): String;
begin
  Result := AString;
  if (Length(AString) > 0) and not (AString[Length(AString)] in ['.', '!', '?']) then
    Result := Result + '.';
end;

{:This function returns the remainder of "AString" after any leading or trailing
spaces are removed.}
function Trim(const AString: string): string;
var
  i, b, e: Integer;
begin
  b := 1;
  e := Length(AString);
  for i := 1 to e do
    if AString[i] = ' ' then
      b := i + 1
    else
      Break;
  for i := Length(AString) downto b do
    if AString[i] = ' ' then
      e := i - 1
    else
      Break;
  if b > e then
    Result := ''
  else
    Result := Copy(AString, b, e - b + 1)
end;

function TrimChars(const AString: string; TrimChars: CharSet): string;
begin
  Result := LTrimChars(RTrimChars(AString, TrimChars), TrimChars);
end;

{:This function convers a spelled-out number from "zero" to "twenty" to an
Integer}
function WordToInt(const Word: String): Integer;
var
	LWord: String;
begin
	LWord := Lowercase(Word);
	if (LWord = 'zero') or (LWord = 'none') then Result := 0
	else if LWord = 'one' then Result := 1
	else if LWord = 'two' then Result := 2
	else if LWord = 'three' then Result := 3
	else if LWord = 'four' then Result := 4
	else if LWord = 'five' then Result := 5
	else if LWord = 'six' then Result := 6
	else if LWord = 'seven' then Result := 7
	else if LWord = 'eight' then Result := 8
	else if LWord = 'nine' then Result := 9
	else if LWord = 'ten' then Result := 10
	else if LWord = 'eleven' then Result := 11
	else if LWord = 'twelve' then Result := 12
	else if LWord = 'thirteen' then Result := 13
	else if LWord = 'fourteen' then Result := 14
	else if LWord = 'fifteen' then Result := 15
	else if LWord = 'sixteen' then Result := 16
	else if LWord = 'seventeen' then Result := 17
	else if LWord = 'eighteen' then Result := 18
	else if LWord = 'nineteen' then Result := 19
	else if LWord = 'twenty' then Result := 20
  else
  	raise EConvertError.Create('Unable to convert word to integer');
end;


{initialization
  TestSplitString;}


end.

