{ VERSION: 1.4 }
unit uCSV;

interface
uses
  Classes, DB;

const
  CsvSeparator = ','; // TUNE: ';'

type
  TCSVFields = class;
  TCSVField = class
  private
    FFieldName: string;
    FDataType: TFieldType;
    FFieldNo: Integer;
    FFields: TCSVFields;
    function GetValue: Variant;
    procedure SetValue(const AValue: Variant);
  public
    property FieldName: string read FFieldName;
    property DataType: TFieldType read FDataType write FDataType;
    property FieldNo: Integer read FFieldNo;
  public
    function IsNull: Boolean;
    procedure Clear;
    property Value: Variant read GetValue write SetValue;
  end;
  TCSVFields = class
  private
    FList: TList;
    FNames: TStrings;
    FCount: Integer;
    FValues: TStrings;
    function GetField(Index: Integer): TCSVField;
    procedure ClearBuffer;
    procedure Clear;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TCSVFields);
    function Add(const Name: string; DataType: TFieldType): TCSVField;
    property Count: Integer read FCount;
    function FieldByName(const FieldName: string): TCSVField;
    function FindField(const FieldName: string): TCSVField;
    property Fields[Index: Integer]: TCSVField read GetField; default;
  end;
  TCSV = class
  private
    FFields: TCSVFields;
    function GetFieldValue(const FieldName: string): Variant;
    procedure SetFieldValue(const FieldName: string; const Value: Variant);
  protected
    FRecordCount: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Format(Structure: TCSVFields);
    property Fields: TCSVFields read FFields;
    property RecordCount: Integer read FRecordCount;
    function FieldCount: Integer;
    function FieldByName(const FieldName: string): TCSVField;
    function FindField(const FieldName: string): TCSVField;
    property FieldValues[const FieldName: string]: Variant read GetFieldValue write SetFieldValue; default;
  end;

  TCSVReader = class(TCSV)
  private
    FLines: TStrings;
    procedure Split(const Line: string; Cols: TStrings);
  public
    constructor Create(const AFileName: string);
    destructor Destroy; override;
  private
    FRecNo: Integer;
  protected
    FEof: Boolean;
  public
    procedure First;
    procedure Next;
    procedure Last;
    property Eof: Boolean read FEof;
  end;

  TCSVWriter = class(TCSV)
  private
    FHeaderWritten: Boolean;
    FStream: TStream;
    procedure WriteHeader;
    procedure Write(Cols: TStrings);
  public
    constructor Create(const AFileName: string);
    destructor Destroy; override;
    procedure Append;
    procedure Post;
  end;

implementation
uses
  DBConsts, Variants, SysUtils;

function Decode(const Str: string; DataType: TFieldType): Variant;
begin
  if Str = '' then begin
    Result := Null;
    exit
  end;

  case DataType of
  ftInteger:
    Result := StrToIntDef(Str, 0);

  ftFloat:
    Result := StrToFloat(Str);

  ftDate:
    Result := StrToDate(Str);

  ftDateTime:
    Result := StrToDateTime(Str);

  ftString:
    if Str[1] = '"' then
      Result := Copy(Str, 2, Length(Str) - 2)
    else
      Result := Str;

  else
    if Str[1] = '"' then
      Result := Decode(Str, ftString)
    else if StrToIntDef(Str, MaxInt) <> MaxInt then
      Result := Decode(Str, ftInteger)
    else
    // TODO: guess type
      Result := Str
  end
end;

function Encode(const Val: Variant; DataType: TFieldType): string;

  function Contains(const Chars: string; const Text: string): Boolean;
  var
    I: Integer;
  begin
    Result := False;
    for I := 1 to Length(Chars) do
      if Pos(Chars[I], Text) <> 0 then begin
        Result := True;
        exit
      end
  end;

begin
  case DataType of
  ftInteger:
    Result := IntToStr(Val);

  ftFloat:
    Result := FloatToStr(Val);

  ftString:
    begin
      Result := Val;
      if Contains(' ,;' { TUNE: }, Result) then
        Result := '"' + Result + '"'
    end;

  else
    Result := VarToStrDef(Val, '')
  end
end;

procedure TCSVField.Clear;
begin
  FFields.FValues[FFieldNo - 1] := ''
end;

function TCSVField.IsNull: Boolean;
begin
  Result := FFields.FValues[FFieldNo - 1] = ''
end;

function TCSVField.GetValue: Variant;
begin
  Result := Decode(FFields.FValues[FFieldNo - 1], FDataType)
end;

procedure TCSVField.SetValue(const AValue: Variant);
begin
  FFields.FValues[FFieldNo - 1] := Encode(AValue, FDataType)
end;

constructor TCSVFields.Create;
begin
  inherited Create;
  FList := TList.Create;
  FNames := TStringList.Create;
  TStringList(FNames).Sorted := True;
  FValues := TStringList.Create
end;

procedure TCSVFields.Clear;
var
  I: Integer;
begin
  for I := Count - 1 downto 0 do
    Fields[I].Free;
  if Assigned(FList) then FList.Clear;
  if Assigned(FNames) then FNames.Clear;
  if Assigned(FValues) then FValues.Clear
end;

destructor TCSVFields.Destroy;
begin
  Clear;
  FList.Free;
  FNames.Free;
  FValues.Free;
  inherited
end;

procedure TCSVFields.Assign(Source: TCSVFields);
var
  I: Integer;
begin
  Clear;
  for I := 0 to Source.Count - 1 do
    Add(Source[I].FieldName, Source[I].DataType)
end;

function TCSVFields.GetField(Index: Integer): TCSVField;
begin
  Result := TCSVField(FList[Index])
end;

function TCSVFields.Add(const Name: string; DataType: TFieldType): TCSVField;
begin
  Result := TCSVField.Create;
  Result.FFields := Self;
  Result.FFieldName := Name;
  Result.FDataType := DataType;
  FList.Add(Result);
  FNames.AddObject(Name, Result);
  FValues.Add('');
  Inc(FCount);
  Result.FFieldNo := FCount
end;

function TCSVFields.FindField(const FieldName: string): TCSVField;
var
  I: Integer;
begin
  I := FNames.IndexOf(FieldName);
  if I < 0 then
    Result := nil
  else
    Result := TCSVField(FNames.Objects[I])
end;

function TCSVFields.FieldByName(const FieldName: string): TCSVField;
begin
  Result := FindField(FieldName);
  if Result = nil then DatabaseErrorFmt(SFieldNotFound, [FieldName]);
end;

procedure TCSVFields.ClearBuffer;
var
  I: Integer;
begin
  for I := FValues.Count - 1 downto 0 do
    FValues[I] := ''
end;

constructor TCSV.Create;
begin
  inherited Create;
  FFields := TCSVFields.Create;
end;

destructor TCSV.Destroy;
begin
  FFields.Free;
  inherited
end;

function TCSV.FieldCount: Integer;
begin
  Result := Fields.Count
end;

function TCSV.FieldByName(const FieldName: string): TCSVField;
begin
  Result := Fields.FieldByName(FieldName)
end;

function TCSV.FindField(const FieldName: string): TCSVField;
begin
  Result := Fields.FindField(FieldName)
end;

function TCSV.GetFieldValue(const FieldName: string): Variant;
begin
  Result := Fields.FieldByName(FieldName).Value
end;

procedure TCSV.SetFieldValue(const FieldName: string; const Value: Variant);
begin
  Fields.FieldByName(FieldName).Value := Value
end;

procedure TCSV.Format(Structure: TCSVFields);
var
  I: Integer;
  F: TCSVField;
begin
  if not Assigned(Structure) then exit;
  for I := 0 to Structure.Count - 1 do begin
    F := FindField(Structure[I].FieldName);
    if Assigned(F) then
      F.DataType := Structure[I].DataType
  end
end;

procedure TCSVReader.Split(const Line: string; Cols: TStrings);
begin
  Cols.Text := StringReplace(Line, CsvSeparator, #13#10, [rfReplaceAll]);
  while Cols.Count < Fields.Count do
    Cols.Add('')
end;

procedure TCSVReader.First;
begin
  FRecNo := -1;
  Next
end;

procedure TCSVReader.Last;
begin
  FRecNo := (RecordCount - 1) - 1;
  Next
end;

procedure TCSVReader.Next;
begin
  Inc(FRecNo);
  if FRecNo >= RecordCount then begin
    FFields.ClearBuffer;
    Dec(FRecNo);
    FEof := True
  end else begin
    FEof := False;
    Split(FLines[FRecNo], FFields.FValues)
  end
end;

constructor TCSVReader.Create(const AFileName: string);
var
  Stream: TStream;
  Cols: TStrings;
  I: Integer;
begin
  inherited Create;
  FLines := TStringList.Create;
  Stream := TFileStream.Create(AFileName, fmOpenRead or fmShareDenyNone);
  try
    FLines.LoadFromStream(Stream);

    if FLines.Count > 0 then begin
      Cols := TStringList.Create;
      try
        Split(FLines[0], Cols);
        for I := 0 to Cols.Count - 1 do
          Fields.Add(Decode(Cols[I], ftString), ftUnknown)
      finally
        Cols.Free
      end;
      FLines.Delete(0);
      FRecordCount := FLines.Count
    end;
    First
  finally
    Stream.Free
  end
end;

destructor TCSVReader.Destroy;
begin
  FLines.Free;
  inherited
end;

constructor TCSVWriter.Create(const AFileName: string);
begin
  inherited Create;
  FStream := TFileStream.Create(AFileName, fmCreate);
end;

destructor TCSVWriter.Destroy;
begin
  if Assigned(FStream) then
    WriteHeader;
  FStream.Free;
  inherited
end;

procedure TCSVWriter.WriteHeader;
var
  I: Integer;
  Cols: TStrings;
begin
  if not FHeaderWritten then begin
    FHeaderWritten := True;
    Cols := TStringList.Create;
    try
      for I := 0 to Fields.Count - 1 do
        Cols.Add(Fields[I].FieldName);
      Write(Cols)
    finally
      Cols.Free
    end
  end
end;

procedure TCSVWriter.Write(Cols: TStrings);
var
  I: Integer;
  Buf: string;
  ColValue: string;
begin
  Buf := '';
  for I := 0 to Cols.Count - 1 do begin
    if I > 0 then
      Buf := Buf + CsvSeparator;
    ColValue := Cols[I];
    if (I = 0) and SameText(ColValue, 'ID') then
      ColValue := '"' + ColValue + '"';
    Buf := Buf + ColValue
  end;
  Buf := Buf + #13#10;
  FStream.WriteBuffer(Buf[1], Length(Buf))
end;

procedure TCSVWriter.Append;
begin
  FFields.ClearBuffer
end;

procedure TCSVWriter.Post;
begin
  WriteHeader;
  Write(FFields.FValues);
  Inc(FRecordCount)
end;

end.
