unit uTodoList;

interface
uses
  fileSys, IniFilesD5, Classes, SysUtils, DB, DBClient;

type
  TGetSortKey = function (Sender: TObject; Data: TDataSet): string of object;
  TCompareKey = function (const A: string; const B: string): Integer of object;
  TDataNavigator = class(TComponent)
  private
    FAscending: Boolean;
    FOnCompareKey: TCompareKey;
    FOnGetSortKey: TGetSortKey;
    FIndex: Integer;
    FEof: Boolean;
    FDataSet: TDataSet;
    FKeys: TList;
    FSortKeys: TStrings;
    FKeyField: string;
    FField: TField;
    procedure SetDataSet(Value: TDataSet);
    procedure SetKeyField(const Value: string);
    procedure SetCurrentIndex(Value: Integer);
  private
    function SortKey: string;
  public
    constructor Create(AnOwner: TComponent); override;
    destructor Destroy; override;
  public
    procedure Clear;
    procedure Add;
    function Remove: Boolean;
    procedure Changed;
    procedure Sort;
  public
    procedure BindFields;
    procedure First;
    procedure Next;
    procedure Prev;
    procedure Last;
    function IsEmpty: Boolean;
    property Eof: Boolean read FEof;
    property Bof: Boolean read FEof;
    property DataSet: TDataSet read FDataSet write SetDataSet;
    property KeyField: string read FKeyField write SetKeyField;
    property OnGetSortKey: TGetSortKey read FOnGetSortKey write FOnGetSortKey;
    property OnCompareKey: TCompareKey read FOnCompareKey write FOnCompareKey;
    property Ascending: Boolean read FAscending write FAscending default True;
  public
    function Count: Integer;
    property CurrentIndex: Integer read FIndex write SetCurrentIndex;
  end;

const
  di_Todos     = 1;
  di_TodoNotes = 2;
type
  TDataState = (dUnknown,dInsert,dEdit,dDelete);
  TNewRecordEvent = procedure (Sender: TDataSet; Updating: Boolean) of object;
  TActivity = (aActive, aComplete, aClosed, aDeleted);
  TTodoList = class(TComponent)
  private
    FOnChange: TNotifyEvent;
    FOnNewRecord: TNewRecordEvent;
    FOnRecordChanged: TDataSetNotifyEvent;
    FOnDeleteRecord: TDataSetNotifyEvent;
    FOnBindFields: TDataSetNotifyEvent;
    FTodoIDField: TField;
    FTodoNotesNavigator: TDataNavigator;
    FModified: Boolean;
    FDataLoaded: Boolean;
    FDataState: array [0..2] of TDataState;
    FLastID: array [0..2] of Integer;
    FNewID: array [0..2] of Integer;
    FKeyField: array [0..2] of string;
    procedure DataChanged(Sender: TDataSet);
    procedure DataInserting(Sender: TDataSet);
    procedure DataDeleting(Sender: TDataSet);
    procedure DataEditing(Sender: TDataSet);
    procedure InitializeRecord(Sender: TDataSet);
    procedure UpdateData(Sender: TDataSet); overload;
    function GetTodoNotesNavigator: TDataNavigator;
  private
    FReadOnly: Boolean;
    FConfig: TCustomIniFile;
    FFiles: TFileSystem;
    FPassword: string;
    procedure SetPassword(const Value: string);
  protected
    FTodos: TDataSet;
    FTodoNotes: TDataSet;
    FFileName: string;
    FFullFileName: string;
    procedure Close;
    procedure DataLoaded;
    procedure UpdateData(Sender: TDataSet; State: TDataState); overload; virtual;
    function GenID(Sender: TDataSet): Integer; virtual;
    procedure Changed; virtual;
    function SolidCache: Boolean; virtual;
  protected
    function TodosKeyField: string; virtual;
    function TodoNotesKeyField: string; virtual;
    function TodoNotesMasterField: string; virtual;
    function TodosMemoField: string; virtual;
    function TodoNotesMemoField: string; virtual;
  public
    procedure AfterConstruction; override;
    destructor Destroy; override;
  public
    function Connect(const AFileName: string; CreateNew: Boolean = False; ForceReadOnly: Boolean = False): Boolean; virtual;
    property FileName: string read FFileName;
    property FullFileName: string read FFullFileName;
    property Files: TFileSystem read FFiles;
    property ReadOnly: Boolean read FReadOnly;
    property Config: TCustomIniFile read FConfig;
    property Password: string read FPassword write SetPassword;
  public
    procedure LoadData; virtual;
    function ValidateUser(const User: string; const Password: string): string; virtual;
    property Todos: TDataSet read FTodos;
    property TodoNotes: TDataSet read FTodoNotes;
    property TodoNotesNavigator: TDataNavigator read GetTodoNotesNavigator;
    property OnNewRecord: TNewRecordEvent read FOnNewRecord write FOnNewRecord;
    property OnRecordChanged: TDataSetNotifyEvent read FOnRecordChanged write FOnRecordChanged;
    property OnDeleteRecord: TDataSetNotifyEvent read FOnDeleteRecord write FOnDeleteRecord;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnBindFields: TDataSetNotifyEvent read FOnBindFields write FOnBindFields;
    property Modified: Boolean read FModified write FModified;
  private
    function TodoID: Integer;
  public
    ActivityEnum,
    CategoryEnum,
    StatusEnum,
    PriorityEnum: TEnum;
  public
    function GetParam(const ID: string; const DefaultValue: string): string;
    procedure SetParam(const ID: string; const Value: string);
  public
    function TodoRecordActivity: TActivity;
    procedure Post(Sender: TDataSet);
  end;

implementation
uses
  uMyskaDB, PackFileSys, RARFileSys;

constructor TDataNavigator.Create(AnOwner: TComponent);
begin
  inherited Create(AnOwner);
  FKeys := TList.Create;
  FSortKeys := TStringList.Create;
  FAscending := True
end;

destructor TDataNavigator.Destroy;
begin
  FKeys.Free;
  FSortKeys.Free;
  inherited
end;

procedure TDataNavigator.BindFields;
begin
  FField := nil
end;

procedure TDataNavigator.SetDataSet(Value: TDataSet);
begin
  FDataSet := Value;
  FField := nil;
  Clear
end;

procedure TDataNavigator.SetKeyField(const Value: string);
begin
  FKeyField := Value;
  FField := nil;
  Clear
end;

procedure TDataNavigator.Clear;
begin
  FKeys.Clear;
  FSortKeys.Clear
end;

function TDataNavigator.SortKey: string;
begin
  if Assigned(FOnGetSortKey) then
    Result := FOnGetSortKey(Self, FDataSet)
  else begin
    if not Assigned(FField) then
      FField := FDataSet.FieldByName(FKeyField);
    Result := IntToStr(FField.AsInteger);
    Result := StringOfChar('0', 6 - Length(Result)) + Result
  end
end;

procedure TDataNavigator.Add;
begin
  if not Assigned(FField) then
    FField := FDataSet.FieldByName(FKeyField);
  FKeys.Add(Pointer(FField.AsInteger));
  FSortKeys.Add(SortKey)
end;

function TDataNavigator.Remove: Boolean;
var
  I: Integer;
begin
  if not Assigned(FField) then
    FField := FDataSet.FieldByName(FKeyField);
  I := FKeys.IndexOf(Pointer(FField.AsInteger));
  if I < 0 then
    Result := False
  else begin
    Result := True;
    FKeys.Delete(I);
    FSortKeys.Delete(I);
  end
end;

procedure TDataNavigator.Changed;
var
  I: Integer;
  NewKey: string;
begin
  if not Assigned(FField) then
    FField := FDataSet.FieldByName(FKeyField);
  I := FKeys.IndexOf(Pointer(FField.AsInteger));
  NewKey := SortKey;
  if FSortKeys[I] <> NewKey then begin
    FSortKeys[I] := NewKey;
    // TODO: use insert sort
    Sort
  end
end;

procedure TDataNavigator.Sort;

  procedure ExchangeItems(I, J: Integer);
  begin
    FKeys.Exchange(I, J);
    FSortKeys.Exchange(I, J);
  end;

  function Compare(I, J: Integer): Integer;
  var
    IKey: string;
    JKey: string;
  begin
    IKey := FSortKeys[I];
    JKey := FSortKeys[J];
    if Assigned(FOnCompareKey) then
      Result := FOnCompareKey(IKey, JKey)
    else begin
      if IKey = JKey then
        Result := 0
      else
      if IKey > JKey then
        Result := 1
      else
        Result := -1
    end;
    if not FAscending then
      Result := -Result
  end;

  procedure QuickSort(L, R: Integer);
  var
    I, J, P: Integer;
  begin
    repeat
      I := L;
      J := R;
      P := (L + R) shr 1;
      repeat
        while Compare(I, P) < 0 do Inc(I);
        while Compare(J, P) > 0 do Dec(J);
        if I <= J then
        begin
          ExchangeItems(I, J);
          if P = I then
            P := J
          else if P = J then
            P := I;
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then QuickSort(L, J);
      L := I;
    until I >= R;
  end;

begin
  if Count > 1 then
    QuickSort(0, Count - 1)
end;

procedure TDataNavigator.First;
begin
  FIndex := -1;
  FEof := False;
  Next
end;

procedure TDataNavigator.SetCurrentIndex(Value: Integer);
begin
  if (Value >= 0) and (Value < FKeys.Count) and FDataSet.Locate(FKeyField, Integer(FKeys[Value]), []) then begin
    FIndex := Value;
    FEof := False
  end else
    FEof := True
end;

procedure TDataNavigator.Next;
begin
  CurrentIndex := CurrentIndex + 1
end;

procedure TDataNavigator.Prev;
begin
  CurrentIndex := CurrentIndex - 1
end;

procedure TDataNavigator.Last;
begin
  CurrentIndex := Count - 1
end;

function TDataNavigator.IsEmpty: Boolean;
begin
  Result := FKeys.Count = 0
end;

function TDataNavigator.Count: Integer;
begin
  Result := FKeys.Count
end;

//---------------------------------------------------------------

procedure TTodoList.Changed;
begin
  FModified := True;
  if Assigned(FOnChange) then
    FOnChange(Self)
end;

procedure TTodoList.Close;

  procedure SaveData(Data: TClientDataSet; const TableName: string);
  var
    Stream: TStream;
  begin
    Stream := Files.GetFileStream(TableName, fmCreate);
    try
      Data.SaveToStream(Stream)
    finally
      Stream.Free
    end
  end;

  procedure WriteEnum(const EnumName: string; Enum: TEnum);
  begin
    if Assigned(Enum) then
      FConfig.WriteString('Enums', EnumName, Enum.CommaText)
  end;

  procedure FreeObjects;
  begin
    FreeAndNil(FTodos);
    FreeAndNil(FTodoNotes);
    FreeAndNil(FTodoNotesNavigator);

    FreeAndNil(FConfig);
  end;

begin
  if Assigned(FFiles) then begin
    if FModified and FDataLoaded then begin
      WriteEnum('Category', CategoryEnum);
      WriteEnum('Status', StatusEnum);
      WriteEnum('Activity', ActivityEnum);
      WriteEnum('Priority', PriorityEnum);

      SaveData(TClientDataSet(Todos), 'Main.dat');
      SaveData(TClientDataSet(TodoNotes), 'MainNotes.dat');

      FConfig.WriteInteger('ID', 'Main', FLastID[di_Todos]);
      FConfig.WriteInteger('ID', 'MainNotes', FLastID[di_TodoNotes]);
    end;

    FreeObjects;

    FFiles.Release;
    FFiles := nil;
    FFileName := '';
    FModified := False;
    FDataLoaded := False;
  end else begin
    FreeObjects;
    FFileName := '';
    FModified := False;
    FDataLoaded := False;
  end
end;

function TTodoList.Connect(const AFileName: string; CreateNew: Boolean; ForceReadOnly: Boolean): Boolean;
var
  BaseFileName: string;
begin
  Close;

  if SameText(ExtractFileExt(AFileName), '.todolist') then begin
    BaseFileName := ChangeFileExt(AFileName, '');
    FFullFileName := AFileName
  end else begin
    BaseFileName := AFileName;
    FFullFileName := AFileName + '.todolist'
  end;

  if (AFileName = '') or ((not CreateNew) and (not FileExists(BaseFileName)) and (not FileExists(FullFileName))) then begin
    Result := False;
    exit;
  end;
  FReadOnly := False;

  if CreateNew then begin
    FFiles := TRARFileSystem.Create(FullFileName, fmCreate, Password)
  end else begin
    if ForceReadOnly then begin
      FFiles := TRARFileSystem.Create(FullFileName, fmOpenRead, Password);
      FReadOnly := True
    end else begin
      try
        FFiles := TRARFileSystem.Create(FullFileName, fmOpenReadWrite, Password);
      except
        on EBadPassword do
          raise;
        on Exception do begin
          if FileExists(FullFileName) then begin
            FFiles := TRARFileSystem.Create(FullFileName, fmOpenRead, Password);
            FReadOnly := True
          end else begin
            try
              FFiles := TRARFileSystem.Create(FullFileName, fmCreate, Password);
            except
              FFiles := TRARFileSystem.Create(FullFileName, fmOpenRead, Password);
              FReadOnly := True
            end
          end
        end
      end;
    end
  end;

  TRARFileSystem(FFiles).CompressLevel := clBest;
  FFiles.Use;

  FConfig := TFileSysIniFile.Create(FFiles, 'Main.ini');

  FFileName := BaseFileName;
  Result := True;

  FTodos := TClientDataSet.Create(Self);
  FTodos.Tag := di_Todos;     // dataset index
  with TClientDataSet(FTodos) do begin
    FieldDefs.Add('ID', ftInteger, 0, True);
    FieldDefs.Add('Module', ftString, 40, False);
    FieldDefs.Add('Created', ftDateTime);
    FieldDefs.Add('Resolved', ftDateTime);
    FieldDefs.Add('Closed', ftDateTime);
    FieldDefs.Add('Note', ftMemo);
    FieldDefs.Add('Priority', ftInteger);
    FieldDefs.Add('Category', ftInteger);
    FieldDefs.Add('Status', ftInteger);
    FieldDefs.Add('Activity', ftInteger);
    FieldDefs.Add('Milestone', ftDate);
    FieldDefs.Add('ReporterID', ftString, 40, False);
    FieldDefs.Add('SolverID', ftString, 40, False);
    IndexDefs.Add('ID', 'ID', [ixPrimary]);
  end;
  TClientDataSet(FTodos).CreateDataSet;

  FTodoNotes := TClientDataSet.Create(Self);
  FTodoNotes.Tag := di_TodoNotes;   // dataset index
  with TClientDataSet(FTodoNotes) do begin
    FieldDefs.Add('ID', ftInteger, 0, True);
    FieldDefs.Add('PID', ftInteger);
    FieldDefs.Add('Created', ftDateTime);
    FieldDefs.Add('Note', ftMemo);
    IndexDefs.Add('ID', 'ID', [ixPrimary]);
    FieldDefs.Add('ReporterID', ftString, 40, False);
    // TODO: sekundarni grupovaci
    // IndexDefs.Add('PID', ');
  end;
  TClientDataSet(FTodoNotes).CreateDataSet;

  FLastID[di_Todos] := FConfig.ReadInteger('ID', 'Main', 0);
  FLastID[di_TodoNotes] := FConfig.ReadInteger('ID', 'MainNotes', 0);

  if CreateNew then
    // force saving
    Changed
end;

procedure TTodoList.AfterConstruction;
begin
  inherited;
  FKeyField[di_Todos] := TodosKeyField;
  FKeyField[di_TodoNotes] := TodoNotesKeyField;
end;

function TTodoList.TodosKeyField: string;
begin
  Result := 'ID'
end;

function TTodoList.TodoNotesKeyField: string;
begin
  Result := 'ID'
end;

function TTodoList.TodoNotesMasterField: string;
begin
  Result := 'PID'
end;

function TTodoList.TodosMemoField: string;
begin
  Result := 'Note'
end;

function TTodoList.TodoNotesMemoField: string;
begin
  Result := 'Note'
end;

function TTodoList.SolidCache: Boolean;
begin
  Result := True
end;

procedure TTodoList.Post(Sender: TDataSet);
var
  Key: Integer;
  OldHandler: TDataSetNotifyEvent;
begin
  if SolidCache or (Sender.State <> dsInsert) then
    Sender.Post
  else begin
    OldHandler := Sender.AfterPost;
    Sender.AfterPost := nil;
    try
      Sender.Post;
      Sender.Close;
      Sender.Open;
      Sender.Locate(FKeyField[Sender.Tag], FNewID[Sender.Tag], []);
      case Sender.Tag of
      di_Todos:
        begin
          FTodoIDField := nil;
        end;

      di_TodoNotes:
        begin
          FTodoNotesNavigator.FField := nil;
        end
      end;
      if Assigned(FOnBindFields) then
        FOnBindFields(Sender);
    finally
      Sender.AfterPost := OldHandler;
      if Assigned(OldHandler) then
        OldHandler(Sender)
    end
  end
end;

procedure TTodoList.SetPassword(const Value: string);
begin
  if Password <> Value then begin
    FPassword := Value;
    if Assigned(FFiles) then begin
      TRARFileSystem(FFiles).Password := Value;
      Changed
    end
  end
end;

function TTodoList.GetParam(const ID: string; const DefaultValue: string): string;
begin
  if Assigned(FConfig) then
    Result := FConfig.ReadString('Params', ID, DefaultValue)
  else
    Result := DefaultValue
end;

procedure TTodoList.SetParam(const ID: string; const Value: string);
begin
  if Assigned(FConfig) and (GetParam(ID, #1) <> Value) then
    FConfig.WriteString('Params', ID, Value)
end;

procedure TTodoList.LoadData;

  procedure LoadData(Data: TDataSet; const TableName: string;
    const Category: string = '';
    const Activity: string = '';
    const Status: string = '';
    const Priority: string = '');

    procedure AssignField(Src, Dst: TField; SrcEnum, DstEnum: TEnum);
    begin
      if Assigned(Dst) and (not Dst.ReadOnly) then begin
        if Src.IsNull then
          Dst.Clear
        else begin
          if Assigned(SrcEnum) and Assigned(DstEnum) then
            Dst.AsInteger := DstEnum.NameToID(SrcEnum.IDToName(Src.AsInteger, ''), 0)
          else
            Dst.Value := Src.Value
        end
      end
    end;

  var
    Stream: TStream;
    Source: TClientDataSet;
    I: Integer;
    S: string;

    LocalCategoryEnum,
    LocalStatusEnum,
    LocalActivityEnum,
    LocalPriorityEnum: TEnum;
    SrcEnums: TList;
    DstEnums: TList;

    function InGroup(const FieldName: string; const Group: string): Boolean;
    begin
      Result := Pos(',' + LowerCase(FieldName) + ',', ',' + LowerCase(Group) + ',') <> 0
    end;

    function EnumText(Enum: TEnum): string;
    begin
      if Assigned(Enum) then
        Result := Enum.CommaText
      else
        Result := ''
    end;

  begin
    Data.Open;
    try
      Stream := Files.GetFileStream(TableName, fmOpenRead);
      try
        Source := TClientDataSet.Create(nil);
        try
          Source.LoadFromStream(Stream);

          LocalCategoryEnum := nil;
          LocalStatusEnum := nil;
          LocalActivityEnum := nil;
          LocalPriorityEnum := nil;

          LocalCategoryEnum := TEnum.Create(Config.ReadString('Enums', 'Category', EnumText(CategoryEnum)));
          if SameText(LocalCategoryEnum.CommaText, EnumText(CategoryEnum)) then
            FreeAndNil(LocalCategoryEnum);

          LocalStatusEnum := TEnum.Create(Config.ReadString('Enums', 'Status', EnumText(StatusEnum)));
          if SameText(LocalStatusEnum.CommaText, EnumText(StatusEnum)) then
            FreeAndNil(LocalStatusEnum);

          LocalActivityEnum := TEnum.Create(Config.ReadString('Enums', 'Activity', EnumText(ActivityEnum)));
          if SameText(LocalActivityEnum.CommaText, EnumText(ActivityEnum)) then
            FreeAndNil(LocalActivityEnum);

          LocalPriorityEnum := TEnum.Create(Config.ReadString('Enums', 'Priority', EnumText(PriorityEnum)));
          if SameText(LocalPriorityEnum.CommaText, EnumText(PriorityEnum)) then
            FreeAndNil(LocalPriorityEnum);

          SrcEnums := TList.Create;
          DstEnums := TList.Create;

          for I := 0 to Source.FieldCount - 1 do begin
            S := Source.Fields[I].FieldName;
            if InGroup(S, Category) then begin
              SrcEnums.Add(LocalCategoryEnum);
              DstEnums.Add(CategoryEnum)
            end else
            if InGroup(S, Status) then begin
              SrcEnums.Add(LocalStatusEnum);
              DstEnums.Add(StatusEnum)
            end else
            if InGroup(S, Activity) then begin
              SrcEnums.Add(LocalActivityEnum);
              DstEnums.Add(ActivityEnum)
            end else
            if InGroup(S, Priority) then begin
              SrcEnums.Add(LocalPriorityEnum);
              DstEnums.Add(PriorityEnum)
            end else begin
              SrcEnums.Add(nil);
              DstEnums.Add(nil)
            end
          end;

          try
            // TODO: don't copy if structured didn't change

            Source.First;
            while not Source.Eof do begin
              Data.Append;
              for I := 0 to Source.FieldCount - 1 do
                AssignField(Source.Fields[I], Data.FindField(Source.Fields[I].FieldName), TEnum(SrcEnums[I]), TEnum(DstEnums[I]));
              Data.Post;

              if Assigned(FOnNewRecord) then
                FOnNewRecord(Data, True);

              Source.Next
            end;
          finally
            SrcEnums.Free;
            DstEnums.Free;

            LocalCategoryEnum.Free;
            LocalStatusEnum.Free;
            LocalActivityEnum.Free;
            LocalPriorityEnum.Free;
          end
        finally
          Source.Free
        end
      finally
        Stream.Free
      end
    except
      // TODO: ignore only stream doesn't exist
    end
  end;

begin
  LoadData(Todos, 'Main.dat', 'Category', 'Activity', 'Status', 'Priority');
  LoadData(TodoNotes, 'MainNotes.dat');

  DataLoaded;
end;

function TTodoList.ValidateUser(const User: string; const Password: string): string;
var
  S: string;
  Users: TStrings;
  I: Integer;
begin
  Result := '';
  S := GetParam('Users', '');
  if S <> '' then begin
    Users := TStringList.Create;
    try
      Users.Text := StringReplace(S, ',', #13#10, [rfReplaceAll]);
      I := Users.IndexOfName(User);
      if I >= 0 then
        if HashPassword(Password) <> Users.ValueFromIndex[I] then
          I := -1;
      if I < 0 then
        raise EBadPassword.Create('')   // TODO: ?
      else
        Result := Users.Names[I]  
    finally
      Users.Free
    end
  end
end;

procedure TTodoList.DataLoaded;

  procedure HookDataSet(Data: TDataSet);
  begin
    Data.BeforeInsert := DataInserting;
    Data.BeforeDelete := DataDeleting;
    Data.BeforeEdit := DataEditing;
    Data.BeforePost := UpdateData;

    Data.AfterPost := DataChanged;
    Data.AfterDelete := DataChanged;

    Data.OnNewRecord := InitializeRecord;
  end;

begin
  FDataLoaded := True;

  HookDataSet(FTodos);
  HookDataSet(FTodoNotes);
end;

procedure TTodoList.UpdateData(Sender: TDataSet; State: TDataState);
begin
end;

procedure TTodoList.UpdateData(Sender: TDataSet);
var
  F: TField;
begin
  if FDataState[Sender.Tag] = dInsert then begin
    F := Sender.FieldByName(FKeyField[Sender.Tag]);
    if F.IsNull or (F.AsInteger = 0) then begin
      F.AsInteger := FNewID[Sender.Tag];
      FLastID[Sender.Tag] := FNewID[Sender.Tag]
    end
  end;
  UpdateData(Sender, FDataState[Sender.Tag])
end;

procedure TTodoList.DataChanged(Sender: TDataSet);
begin
  Changed;
  case FDataState[Sender.Tag] of
  dInsert:
    if Assigned(FOnNewRecord) then
      FOnNewRecord(Sender, False);

  dEdit:
    if Assigned(FOnRecordChanged) then
      FOnRecordChanged(Sender);

  dDelete:
    ;
  end;
  if (Sender = FTodoNotes) and Assigned(FTodoNotesNavigator) then
    // invalidate navigator
    // TODO: use smart refresh
    FTodoNotesNavigator.Tag := 0
end;

function TTodoList.TodoID: Integer;
begin
  if not Assigned(FTodoIDField) then
    FTodoIDField := FTodos.FieldByName(TodosKeyField);
  Result := FTodoIDField.AsInteger
end;

function TTodoList.GetTodoNotesNavigator: TDataNavigator;
var
  ID: Integer;
begin
  if not Assigned(FTodoNotesNavigator) then begin
    FTodoNotesNavigator := TDataNavigator.Create(nil);
    FTodoNotesNavigator.DataSet := FTodoNotes;
    FTodoNotesNavigator.KeyField := TodoNotesKeyField
  end;
  ID := TodoID;
  if FTodoNotesNavigator.Tag <> ID then begin
    FTodoNotesNavigator.Clear;
    FTodoNotesNavigator.Tag := ID;

    FTodoNotes.First;
    while not FTodoNotes.Eof do begin
      if FTodoNotes.FieldValues[TodoNotesMasterField] = ID then
        FTodoNotesNavigator.Add;
      FTodoNotes.Next
    end;
    FTodoNotesNavigator.Sort
  end;
  Result := FTodoNotesNavigator
end;

procedure TTodoList.InitializeRecord(Sender: TDataSet);
begin
  if Sender = FTodoNotes then
    FTodoNotes.FieldValues[TodoNotesMasterField] := TodoID
end;

function TTodoList.GenID(Sender: TDataSet): Integer;
var
  N: Integer;
  F: TField;
begin
  Result := FLastID[Sender.Tag];
  if Result = 0 then begin
    // naskocime na posledni maximum
    F := Sender.FieldByName(FKeyField[Sender.Tag]);
    Sender.First;
    while not Sender.Eof do begin
      N := F.AsInteger;
      if N > Result then
        Result := N;
      Sender.Next
    end;
    Inc(Result)
  end else begin
    while Sender.Locate(FKeyField[Sender.Tag], Result, []) do
      Inc(Result);
  end
end;

procedure TTodoList.DataInserting(Sender: TDataSet);
begin
  FDataState[Sender.Tag] := dInsert;
  FNewID[Sender.Tag] := GenID(Sender)
end;

procedure TTodoList.DataDeleting(Sender: TDataSet);
var
  Nav: TDataNavigator;
begin
  FDataState[Sender.Tag] := dDelete;
  if Assigned(FOnDeleteRecord) then
    FOnDeleteRecord(Sender);
  if Sender = FTodos then begin
    Nav := TodoNotesNavigator;
    while not Nav.IsEmpty do begin
      Nav.First;
      Nav.Remove;
      Nav.DataSet.Delete
    end
  end
end;

procedure TTodoList.DataEditing(Sender: TDataSet);
begin
  FDataState[Sender.Tag] := dEdit;
end;

destructor TTodoList.Destroy;
begin
  Close;
  inherited
end;

//---------------------------------------------------------------------------

function TTodoList.TodoRecordActivity: TActivity;
var
  F: TField;
  S: string;
begin
  F := Todos.FieldByName('Activity');
  Result := aActive;
  if not F.IsNull then begin
    S := ActivityEnum.IDToName(F.AsInteger, ActivityEnum[0]{ TODO: NullActivity});
    if SameText(S, 'complete') then
      Result := aComplete
    else
    if SameText(S, 'closed') then
      Result := aClosed
    else
    if SameText(S, 'deleted') then
      Result := aDeleted
  end
end;

end.
