{ VERSION: 1.1 }
{ 1.1 / 4.6.2006
  - new event OnSetStyle to enable enforcement of a style logic
}
unit uWebServer;

interface
uses
  Classes, HttpSrv, uWebUI;

type
  TValidProjectName = function (const ProjectName: string): Boolean of object;
  TOnSetClient = procedure (const Host: string; const User: string; const Password: string) of object;
  TWebServer = class(TComponent)
  private
    FValidProjectName: TValidProjectName;
    FOnSetClient: TOnSetClient;
    FOnSetStyle: TNotifyEvent;
    FPager: TPageGenerator;
    FAutoPort: Boolean;
    FPort: Integer;
    FActive: Boolean;
    FDefaultPage: string;
    FDocDir: string;
    FWebRootFolder: string;
    FOnStart: TNotifyEvent;
    FOnStop: TNotifyEvent;
    procedure SetActive(Value: Boolean);
    function GetPort: Integer;
    procedure InternalGetDocument(Sender, Client: TObject; var Flags: THttpGetFlag);
    procedure InternalPostDocument(Sender, Client: TObject; var Flags: THttpGetFlag);
    procedure InternalPostedData(Sender, Client: TObject; Error: Word);
  protected
    Handle: THttpServer;
    procedure DoConnect; virtual;
    procedure DoDisconnect; virtual;
    procedure DoGetDocument(Client: THttpConnection; var Flags: THttpGetFlag); virtual;
  public
    constructor Create(AnOwner: TComponent); override;
    destructor Destroy; override;
  public
    procedure Start;
    procedure Stop;
    property Port: Integer read GetPort write FPort default 8081;
    property AutoPort: Boolean read FAutoPort write FAutoPort;
    property DefaultPage: string read FDefaultPage write FDefaultPage;
    property DocDir: string read FDocDir write FDocDir;
    property WebRootFolder: string read FWebRootFolder write FWebRootFolder;
    property Active: Boolean read FActive write SetActive;
    property OnStart: TNotifyEvent read FOnStart write FOnStart;
    property OnStop: TNotifyEvent read FOnStop write FOnStop;
    property Pager: TPageGenerator read FPager write FPager;
    property ValidProjectName: TValidProjectName read FValidProjectName write FValidProjectName;
    property OnSetClient: TOnSetClient read FOnSetClient write FOnSetClient;
    property OnSetStyle: TNotifyEvent read FOnSetStyle write FOnSetStyle;
  end;

implementation
uses
  PackFileSys, HTTPApp, SynaCode, SysUtils, StrUtils;

constructor TWebServer.Create(AnOwner: TComponent);
begin
  FPort := 8081;
  FDefaultPage := 'index.htm';
  inherited Create(AnOwner)
end;

destructor TWebServer.Destroy;
begin
  Active := False;
  inherited
end;

procedure TWebServer.SetActive(Value: Boolean);
begin
  if Active = Value then exit;
  if Value then begin
    DoConnect;
    FActive := True
  end else begin
    DoDisconnect;
    if (not (csDestroying in ComponentState)) and Assigned(FOnStop) then
      FOnStop(Self);
    FActive := False
  end
end;

procedure TWebServer.Start;
begin
  Active := True
end;

procedure TWebServer.Stop;
begin
  Active := False
end;

procedure TWebServer.DoConnect;
begin
  Handle := THttpServer.Create(Self);
  Handle.Port := IntToStr(Port);
  Handle.DefaultDoc := DefaultPage;
  if DocDir <> '' then
    Handle.DocDir := DocDir;
  Handle.OnServerStarted := OnStart;
  Handle.OnGetDocument := InternalGetDocument;
{ TODO:
  Handle.OnPostDocument := InternalPostDocument;
  Handle.OnPostedData := InternalPostedData;
}
  try
    try
      Handle.Start;
    except
      if (Port <> 0) and AutoPort then begin
        Handle.Port := '0';
        Handle.Start
      end else
        raise
    end;
  except
    FreeAndNil(Handle);
    raise
  end
end;

function TWebServer.GetPort: Integer;
begin
  if Assigned(Handle) then
    Result := StrToIntDef(Handle.WSocketServer.GetXPort, FPort)
  else
    Result := FPort
end;

procedure TWebServer.DoDisconnect;
begin
  Handle.Stop;
  FreeAndNil(Handle)
end;

procedure TWebServer.InternalGetDocument(Sender, Client: TObject; var Flags: THttpGetFlag);
begin
  DoGetDocument(THttpConnection(Client), Flags)
end;

type THC = class(THttpConnection) end;
procedure TWebServer.InternalPostDocument(Sender, Client: TObject; var Flags: THttpGetFlag);
var
  Conn: THC absolute Client;
begin
  if SameText(RightStr(Conn.Path, 4), '.exe') and (Conn.RequestContentLength > 0) then
    Flags := hgAcceptData
end;

procedure TWebServer.InternalPostedData(Sender, Client: TObject; Error: Word);
var
  Conn: THC absolute Client;
  Flags: THttpGetFlag;
begin
  Conn.Params := Conn.Params + Conn.ReceiveStr;
  if Length(Conn.Params) >= Conn.RequestContentLength then begin
    Flags := hg404;
    DoGetDocument(Conn, Flags);
    case Flags of
    hg401:
        begin
            Conn.Answer401;
            Conn.CloseDelayed;
        end;
    hg403:
        begin
            Conn.Answer403;
            Conn.CloseDelayed;
        end;
    hg404:
        begin
            Conn.Answer404;
            Conn.CloseDelayed;
        end;
    else
        Conn.CloseDelayed;
    end;
  end
end;

procedure TWebServer.DoGetDocument(Client: THttpConnection; var Flags: THttpGetFlag);

  procedure Answer302(const URL: string);
  var
    Body: string;
  begin
    Body := '<HTML><HEAD><TITLE>302 Moved Temporarily</TITLE></HEAD>' +
            '<BODY><H1>302 Moved Temporarily</H1></BODY></HTML>' + #13#10;
    Client.SendHeader(
      Client.Version + ' 302 Moved Temporarily' + #13#10 +
      'Content-Type: text/html' + #13#10 +
      'Location: ' + URL + #13#10 +
      'Content-Length: ' + IntToStr(Length(Body)) + #13#10 +
      #13#10);
     { Do not use AnswerString method because we don't want to use ranges }
    Client.SendStr(Body);
  end;


  //---------------------------------
  {

    server structure:

    /about.htm - help file
    /[index.htm] - list of projects
    /<project>/todo/[index.htm] - list of todos
    /<project>/todo/history.htm - list of closed todos
    /<project>/todo/resolved.htm - list of resolved but not closed todos
    /<project>/todo/0[/index.htm] - record insert
    /<project>/todo/0/bug.htm - insert new bug
    /<project>/todo/0/task.htm - insert new task
    /<project>/todo/0/feature.htm - insert new feature
    /<project>/todo/0/question.htm - insert new question
    /<project>/todo/<recid>[/index.htm] - record view
    /<project>/todo/<recid>/edit.htm - record editor
    /<project>/todo/<recid>/resolve.htm - record editor
    /<project>/todo/<recid>/close.htm - record editor
    /<project>/todo/<recid>/resolveclose.htm - record editor (resolve and close)
    /<project>/todo/<recid>/reactivate.htm - record editor
    /<project>/todo/<recid>/reopen.htm - record editor
    /<project>/todo/<recid>/delete.htm - record deletor
    /<project>/todo/update.exe?.. - record insert/update

  }

  function IsFolder(PathItems: TStrings; const Path: string): Boolean;

    function IsNumber(const Value: string): Boolean;
    begin
      Result := StrToIntDef(Value, -1) >= 0
    end;

  begin
    Result := True;
    if Path = '' then exit;
  //  if Path = 'about.htm' then exit;
    if Assigned(ValidProjectName) then
      Result := ValidProjectName(UnWebName(PathItems[0]))
    else
      Result := True;
    if not Result then exit;
    if (PathItems.Count < 2) or (not SameText(PathItems[1], 'todo')) then begin
      Result := False;
      exit
    end;
    if PathItems.Count = 2 then exit;
    Result := IsNumber(PathItems[2]);
    if not Result then exit;
    Result := PathItems.Count = 3
  end;

var
  PathItems: TStrings;
  Path: string;
  QueryFields: TStrings;

  procedure AuthDecode(const Auth: string; var User: string; var Password: string);
  const
    Prefix = 'Basic ';
  var
    S: string;
    I: Integer;
  begin
    if SameText(Copy(Auth, 1, Length(Prefix)), Prefix) then begin
      S := SynaCode.DecodeBase64(Copy(Auth, Length(Prefix) + 1, MaxInt));
      I := Pos(':', S);
      if I = 0 then begin
        User := '';
        Password := ''
      end else begin
        User := Copy(S, 1, I - 1);
        Password := Copy(S, I + 1, MaxInt)
      end
    end else begin
      User := '';
      Password := ''
    end
  end;

var
  User: string;
  Password: string;
  OldUserName: string;
begin
  PathItems := TStringList.Create;
  QueryFields := TStringList.Create;
  try
    PathItems.Text := Trim(StringReplace(LowerCase(Client.Path), '/', #13#10, [rfReplaceAll]));
    Path := PathItems.CommaText;

    if WebRootFolder <> '' then
      if (Path <> '') and SameText(PathItems[0], WebRootFolder) then begin
        PathItems.Delete(0);
        Path := PathItems.CommaText
      end else
        exit;

    if IsFolder(PathItems, Path) then begin
      // TODO: do proper HTTP redirect
      PathItems.Add(DefaultPage);
      Path := PathItems.CommaText
    end;
    ExtractHTTPFields(['&'], [], PChar(Client.Params), QueryFields);
    AuthDecode(Client.RequestAuth, User, Password);

    if Client.RequestUserAgent = 'Todo' then begin
      Pager.Style := [psBlankProject, psFixedFontSize]
    end else begin
      Pager.Style := [psFixedFontSize, psProjectList];
    end;

    if Assigned(FOnSetStyle) then
      FOnSetStyle(Self);

    if Pager.FontSize = 0 then
      Pager.Style := Pager.Style - [psFixedFontSize];

    try
      if Assigned(FOnSetClient) then
        FOnSetClient(Client.RequestHost, User, Password);
      OldUserName := Pager.UserName;  
      try
        Client.AnswerString(Flags, '', '', 'Pragma: no-cache'#13#10, Pager.GetPage(PathItems, Path, QueryFields, User, Password));
      finally
        Pager.UserName := OldUserName;
        if Assigned(FOnSetClient) then
          FOnSetClient('', '', '')
      end
    except
      on EBadPassword do begin
        Flags := hg401;
        THC(Client).FBasicRealm := PathItems[0];
      end;
      on e:ERedirect do begin
        Flags := hgWillSendMyself;
        Answer302(e.Message)
      end;
      on Exception do begin
        // TODO: throw 500 - internal error
        Flags := hg404;
      end
    end;
  finally
    QueryFields.Free;
    PathItems.Free
  end
end;

end.
